--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity SDREADIF is
	port (
		SD_DAT0		: in  std_logic;						-- from SD card
		SD_WPN		: in  std_logic;						-- from SD card
		ADD			: in  std_logic_vector(9 downto 0);		-- from CPU
		DATAI		: in  std_logic_vector(7 downto 0);		-- from CPU
		WRN			: in  std_logic;						-- from CPU
		RDN			: in  std_logic;						-- from CPU
		CMTRDOPEN	: in  std_logic;						-- from 8049
		CMTRDREQ	: in  std_logic;						-- from 8049
		CMTRDACC	: in  std_logic;						-- from 8049
		CMTRDACCMD	: in  std_logic;
		CMTRDACCHMD	: in  std_logic;
		CMTWROPEN	: in  std_logic;						-- from 8049
		CMTWRREQ	: in  std_logic;						-- from 8049
		CMTWRDT		: in  std_logic_vector(7 downto 0);		-- from 8049
		FPGAVER		: in  std_logic_vector(15 downto 0);
		CLK50M		: in  std_logic;
		RSTN		: in  std_logic;
		DATAO		: out std_logic_vector(7 downto 0);		-- to CPU
		SD_RDAD		: out std_logic_vector(18 downto 0);	-- to SDRAM/CGROM
		SD_RDDT		: out std_logic_vector(7 downto 0);		-- to SDRAM/CGROM
		SD_SDENB	: out std_logic;						-- to SDRAM
		SD_SDWRN	: out std_logic;						-- to SDRAM
		SD_CGENB	: out std_logic;						-- to CGROM
		SD_CGWRN	: out std_logic;						-- to CGROM
		SD_RDDONE	: out std_logic;
		SD_ERR		: out std_logic;
		SD_OUTENB	: out std_logic;
		CMTRDY		: out std_logic;						-- to 8049
		CMTRDDT		: out std_logic_vector(7 downto 0);		-- to 8049
		CMTCNT		: out std_logic_vector(15 downto 0);
		FIRMVER		: out std_logic_vector(15 downto 0);
		SD_CMD		: out std_logic;						-- to SD card
		SD_DAT		: out std_logic_vector(3 downto 1);		-- to SD card
		SD_CLK		: out std_logic							-- to SD card
	);
end SDREADIF;

architecture RTL of SDREADIF is

	component SDREAD_CMDCNT is
		port (
			SD_DAT0		: in  std_logic;
			SD_WPN		: in  std_logic;
			CMD_START	: in  std_logic;
			CMD_NO		: in  std_logic_vector(5 downto 0);
			READ_SDADD	: in  std_logic_vector(31 downto 0);
			SDCNTENB	: in  std_logic;
			CLK			: in  std_logic;
			RSTN		: in  std_logic;
			STATEOUT	: out std_logic_vector(7 downto 0);
			INDT		: out std_logic_vector(7 downto 0);
			INDTENB		: out std_logic;
			INDTLT		: out std_logic;
			INDTLT2		: out std_logic;
			CMD_END		: out std_logic;
			IDLEDET		: out std_logic;
			ERRDET		: out std_logic;
			SD_CMD		: out std_logic;
			SD_DAT		: out std_logic_vector(3 downto 1);
			SD_CLK		: out std_logic
		);
	end component;

	component SPRAM_512W8B is
		port (
			address	: in  std_logic_vector(8 downto 0);
			data	: in  std_logic_vector(7 downto 0);
			wren	: in  std_logic;
			clock	: in  std_logic;
			aclr	: in  std_logic;
			q		: out std_logic_vector(7 downto 0)
		);
	end component;

	signal indt_i		: std_logic_vector(7 downto 0);
	signal indtenb_i	: std_logic;
	signal indtlt2_i	: std_logic;
	signal sd_cnt		: std_logic_vector(8 downto 0);
	signal cmd_end_i	: std_logic;
	signal idledet_i	: std_logic;
	signal errdet_i		: std_logic;
	signal cmd_end_f1	: std_logic;
	signal cmd_end_f2	: std_logic;

	signal fifo_cnt		: std_logic_vector(4 downto 0);
	signal fifo_add		: std_logic_vector(31 downto 0);

	signal cmtrdrq_f1	: std_logic;
	signal cmtrdrq_f2	: std_logic;
	signal cmtrdrq_f3	: std_logic;
	signal cmtrdycnt	: std_logic_vector(20 downto 0);
	signal cmtrdy_i		: std_logic;
	signal cmtrddt_i	: std_logic_vector(7 downto 0);

	signal aclr		: std_logic;
	signal buf1_add	: std_logic_vector(8 downto 0);
	signal buf2_add	: std_logic_vector(8 downto 0);
	signal buf1_dat	: std_logic_vector(7 downto 0);
	signal buf2_dat	: std_logic_vector(7 downto 0);
	signal buf1_wr	: std_logic;
	signal buf2_wr	: std_logic;
	signal buf1_q	: std_logic_vector(7 downto 0);
	signal buf2_q	: std_logic_vector(7 downto 0);
	signal buf_q	: std_logic_vector(7 downto 0);

	signal reg_start		: std_logic;						-- 0x8200(0)
	signal reg_startend		: std_logic;						-- 0x8200(0)
	signal reg_idledet		: std_logic;						-- 0x8200(1)
	signal reg_errdet		: std_logic;						-- 0x8200(2)
	signal reg_inuse		: std_logic;						-- 0x8201(0)
	signal reg_sderr		: std_logic;						-- 0x8201(1)
	signal reg_sdrddone		: std_logic;						-- 0x8201(2)
	signal reg_rominit		: std_logic;						-- 0x8201(3)
	signal reg_rominitend	: std_logic;						-- 0x8201(3)
	signal reg_cmdno		: std_logic_vector(5 downto 0);		-- 0x8202(5:0)
	signal reg_sdadd		: std_logic_vector(31 downto 0);	-- 0x8203-0x8206
	signal reg_firmver		: std_logic_vector(15 downto 0);	-- 0x820E-0x820F
	signal reg_bufs			: std_logic;						-- 0x8210(0)
	signal reg_bufrd		: std_logic;						-- 0x8211(0)
	signal reg_bufinit		: std_logic;						-- 0x8211(1)
	signal reg_bufrdreq		: std_logic;						-- 0x8212(0)
	signal reg_fifordind	: std_logic;						-- 0x8213(0)
	signal reg_fifoinit		: std_logic;						-- 0x8213(1)
	signal reg_sdram		: std_logic;						-- 0x8214(0)
	signal reg_cgrom		: std_logic;						-- 0x8214(1)
	signal reg_cmtrd		: std_logic;						-- 0x8214(2)
	signal reg_cmtwr		: std_logic;						-- 0x8214(3)
	signal reg_fifowrreq	: std_logic;						-- 0x8215(0)
	signal reg_fifordend	: std_logic;						-- 0x8215(1)
	signal reg_romstart		: std_logic_vector(31 downto 0);	-- 0x8216-0x8219
	signal reg_cmtlen		: std_logic_vector(31 downto 0);	-- 0x821A-0x821D
	signal reg_cmtwnom		: std_logic_vector(31 downto 0);	-- 0x8220-0x8223
	signal reg_cmtwaccl		: std_logic_vector(31 downto 0);	-- 0x8224-0x8227
	signal reg_cmtwacch		: std_logic_vector(31 downto 0);	-- 0x8228-0x822B
	signal reg_cmtwst		: std_logic_vector(31 downto 0);	-- 0x822C-0x822F

	signal reg_start_f1	: std_logic;
	signal reg_start_f2	: std_logic;
	signal reg_start_f3	: std_logic;
	signal reg_start_r	: std_logic;

	signal reg_startlt1	: std_logic;
	signal reg_startlt2	: std_logic;

	signal reg_rominit_f1	: std_logic;
	signal reg_rominit_f2	: std_logic;
	signal reg_rominit_f3	: std_logic;
	signal reg_rominit_r	: std_logic;
	signal rominitcnt		: std_logic_vector(17 downto 0);

	signal cmtrdopen_f1	: std_logic;
	signal cmtrdopen_f2	: std_logic;
	signal cmtrdreq_f1	: std_logic;
	signal cmtrdreq_f2	: std_logic;
	signal cmtwropen_f1	: std_logic;
	signal cmtwropen_f2	: std_logic;
	signal cmtwrreq_f1	: std_logic;
	signal cmtwrreq_f2	: std_logic;
	signal cmtwrdt_f1	: std_logic_vector(7 downto 0);
	signal cmtwrdt_f2	: std_logic_vector(7 downto 0);

	signal datao_i		: std_logic_vector(7 downto 0);
	signal sd_rdad_i	: std_logic_vector(18 downto 0);
	signal sd_rddt_i	: std_logic_vector(7 downto 0);
	signal sd_sdenb_i	: std_logic;
	signal sd_sdwrn_i	: std_logic;
	signal sd_cgenb_i	: std_logic;
	signal sd_cgwrn_i	: std_logic;
	signal read_sdadd_i	: std_logic_vector(31 downto 0);

	signal rdn_f1	: std_logic;
	signal rdn_f2	: std_logic;
	signal rdn_f3	: std_logic;
	signal wrn_f1	: std_logic;
	signal wrn_f2	: std_logic;
	signal wrn_f3	: std_logic;
	signal add_f1	: std_logic_vector(9 downto 0);
	signal add_f2	: std_logic_vector(9 downto 0);
	signal datai_f1	: std_logic_vector(7 downto 0);
	signal datai_f2	: std_logic_vector(7 downto 0);
	signal rdp		: std_logic;
	signal wrp		: std_logic;

begin

	U_SDREAD_CMDCNT : SDREAD_CMDCNT
	port map (
		SD_DAT0		=> SD_DAT0,
		SD_WPN		=> SD_WPN,
		CMD_START	=> reg_startlt2,
		CMD_NO		=> reg_cmdno,
		READ_SDADD	=> read_sdadd_i,
		SDCNTENB	=> reg_inuse,
		CLK			=> CLK50M,
		RSTN		=> RSTN,
		STATEOUT	=> open,
		INDT		=> indt_i,
		INDTENB		=> indtenb_i,
		INDTLT2		=> indtlt2_i,
		CMD_END		=> cmd_end_i,
		IDLEDET		=> idledet_i,
		ERRDET		=> errdet_i,
		SD_CMD		=> SD_CMD,
		SD_DAT		=> SD_DAT,
		SD_CLK		=> SD_CLK
	);

	read_sdadd_i <= reg_sdadd(22 downto 0) & "000000000";

	U_SPRAM_512W8B_1 : SPRAM_512W8B
	port map (
		address	=> buf1_add,
		data	=> buf1_dat,
		wren	=> buf1_wr,
		clock	=> CLK50M,
		aclr	=> aclr,
		q		=> buf1_q
	);

	U_SPRAM_512W8B_2 : SPRAM_512W8B
	port map (
		address	=> buf2_add,
		data	=> buf2_dat,
		wren	=> buf2_wr,
		clock	=> CLK50M,
		aclr	=> aclr,
		q		=> buf2_q
	);

	aclr     <= not RSTN;

	buf1_add <=	sd_cnt             when (reg_bufs = '0' and reg_bufrd = '0') else
				add_f2(8 downto 0) when (reg_bufs = '0' and reg_bufrd = '1') else
				fifo_add(8 downto 0);
	buf2_add <=	sd_cnt             when (reg_bufs = '1' and reg_bufrd = '0') else
				add_f2(8 downto 0) when (reg_bufs = '1' and reg_bufrd = '1') else
				fifo_add(8 downto 0);

	buf1_dat <= X"00"  when (reg_bufs = '1') else indt_i;
	buf2_dat <= X"00"  when (reg_bufs = '0') else indt_i;
	buf1_wr  <= '0'    when (reg_bufs = '1') else (indtenb_i and indtlt2_i);
	buf2_wr  <= '0'    when (reg_bufs = '0') else (indtenb_i and indtlt2_i);

	buf_q    <= buf1_q when (reg_bufs = '0') else buf2_q;



-- CPU I/F
-- data latch
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			rdn_f1   <= '1';
			rdn_f2   <= '1';
			rdn_f3   <= '1';
			wrn_f1   <= '1';
			wrn_f2   <= '1';
			wrn_f3   <= '1';
			add_f1   <= (others => '0');
			add_f2   <= (others => '0');
			datai_f1 <= (others => '0');
			datai_f2 <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			rdn_f1   <= RDN;
			rdn_f2   <= rdn_f1;
			rdn_f3   <= rdn_f2;
			wrn_f1   <= WRN;
			wrn_f2   <= wrn_f1;
			wrn_f3   <= wrn_f2;
			add_f1   <= ADD;
			add_f2   <= add_f1;
			datai_f1 <= DATAI;
			datai_f2 <= datai_f1;
		end if;
	end process;

	rdp <= '1' when (rdn_f2 = '0' and rdn_f3 = '1') else '0';
	wrp <= '1' when (wrn_f2 = '0' and wrn_f3 = '1') else '0';

-- register
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_start    <= '0';
			reg_inuse    <= '0';
			reg_sderr    <= '0';
			reg_sdrddone <= '0';
			reg_rominit  <= '0';
			reg_cmdno    <= (others => '0');
			reg_sdadd    <= (others => '0');
			reg_firmver  <= (others => '0');
			reg_bufs     <= '0';
			reg_bufrd    <= '0';
			reg_bufinit  <= '0';
			reg_fifordind<= '0';
			reg_sdram    <= '0';
			reg_cgrom    <= '0';
			reg_cmtrd    <= '0';
			reg_cmtwr    <= '0';
			reg_fifoinit <= '0';
			reg_romstart <= (others => '0');
			reg_cmtlen   <= (others => '0');
			reg_cmtwnom  <= X"0007EEEE";
			reg_cmtwaccl <= X"00010000";
			reg_cmtwacch <= X"00010000";
			reg_cmtwst   <= X"00100000";
		elsif (CLK50M'event and CLK50M = '1') then
			if (wrp = '1') then
				case add_f2 is
					when "1000000000"=> reg_start                  <= datai_f2(0);
					when "1000000001"=>	reg_inuse                  <= datai_f2(0);
										reg_sderr                  <= datai_f2(1);
										reg_sdrddone               <= datai_f2(2);
										reg_rominit                <= datai_f2(3);
					when "1000000010"=> reg_cmdno                  <= datai_f2(5 downto 0);
					when "1000000011"=> reg_sdadd( 7 downto  0)    <= datai_f2;
					when "1000000100"=> reg_sdadd(15 downto  8)    <= datai_f2;
					when "1000000101"=> reg_sdadd(23 downto 16)    <= datai_f2;
					when "1000000110"=> reg_sdadd(31 downto 24)    <= datai_f2;
					when "1000001110"=> reg_firmver( 7 downto  0)  <= datai_f2;
					when "1000001111"=> reg_firmver(15 downto  8)  <= datai_f2;
					when "1000010000"=> reg_bufs                   <= datai_f2(0);
					when "1000010001"=> reg_bufrd                  <= datai_f2(0);
										reg_bufinit                <= datai_f2(1);
					when "1000010011"=> reg_fifordind              <= datai_f2(0);
										reg_fifoinit               <= datai_f2(1);
					when "1000010100"=> reg_sdram                  <= datai_f2(0);
										reg_cgrom                  <= datai_f2(1);
										reg_cmtrd                  <= datai_f2(2);
										reg_cmtwr                  <= datai_f2(3);
					when "1000010110"=> reg_romstart( 7 downto  0) <= datai_f2;
					when "1000010111"=> reg_romstart(15 downto  8) <= datai_f2;
					when "1000011000"=> reg_romstart(23 downto 16) <= datai_f2;
					when "1000011001"=> reg_romstart(31 downto 24) <= datai_f2;
					when "1000011010"=> reg_cmtlen( 7 downto  0)   <= datai_f2;
					when "1000011011"=> reg_cmtlen(15 downto  8)   <= datai_f2;
					when "1000011100"=> reg_cmtlen(23 downto 16)   <= datai_f2;
					when "1000011101"=> reg_cmtlen(31 downto 24)   <= datai_f2;
					when "1000100000"=> reg_cmtwnom( 7 downto  0)  <= datai_f2;
					when "1000100001"=> reg_cmtwnom(15 downto  8)  <= datai_f2;
					when "1000100010"=> reg_cmtwnom(23 downto 16)  <= datai_f2;
					when "1000100011"=> reg_cmtwnom(31 downto 24)  <= datai_f2;
					when "1000100100"=> reg_cmtwaccl( 7 downto  0) <= datai_f2;
					when "1000100101"=> reg_cmtwaccl(15 downto  8) <= datai_f2;
					when "1000100110"=> reg_cmtwaccl(23 downto 16) <= datai_f2;
					when "1000100111"=> reg_cmtwaccl(31 downto 24) <= datai_f2;
					when "1000101000"=> reg_cmtwacch( 7 downto  0) <= datai_f2;
					when "1000101001"=> reg_cmtwacch(15 downto  8) <= datai_f2;
					when "1000101010"=> reg_cmtwacch(23 downto 16) <= datai_f2;
					when "1000101011"=> reg_cmtwacch(31 downto 24) <= datai_f2;
					when "1000101100"=> reg_cmtwst( 7 downto  0)   <= datai_f2;
					when "1000101101"=> reg_cmtwst(15 downto  8)   <= datai_f2;
					when "1000101110"=> reg_cmtwst(23 downto 16)   <= datai_f2;
					when "1000101111"=> reg_cmtwst(31 downto 24)   <= datai_f2;
					when others      => null;
				end case;
			end if;
		end if;
	end process;

-- start_end register
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			cmd_end_f1 <= '0';
			cmd_end_f2 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			cmd_end_f1 <= cmd_end_i;
			cmd_end_f2 <= cmd_end_f1;
		end if;
	end process;

	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_start_f1 <= '0';
			reg_start_f2 <= '0';
			reg_start_f3 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			reg_start_f1 <= reg_start;
			reg_start_f2 <= reg_start_f1;
			reg_start_f3 <= reg_start_f2;
		end if;
	end process;

	reg_start_r <= '1' when (reg_start_f2 = '1' and reg_start_f3 = '0') else '0';

	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_startlt1 <= '0';
			reg_startlt2 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_start_r = '1') then
				reg_startlt1 <= '1';
			elsif (indtlt2_i = '1') then
				reg_startlt1 <= '0';
			end if;

			if (indtlt2_i = '1') then
				reg_startlt2 <= reg_startlt1;
			end if;
		end if;
	end process;

	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_startend <= '0';
			reg_idledet  <= '0';
			reg_errdet   <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (cmd_end_f1 = '1' and cmd_end_f2 = '0') then
				reg_startend <= '0';
			elsif (reg_start_r = '1') then
				reg_startend <= '1';
			end if;
			if (cmd_end_f1 = '1' and cmd_end_f2 = '0') then
				reg_idledet  <= idledet_i;
				reg_errdet   <= errdet_i;
			end if;
		end if;
	end process;

-- register read
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			datao_i <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			if (rdp = '1') then
				case add_f2 is
					when "1000000000"=> datao_i <= "00000" & reg_errdet & reg_idledet & reg_startend;
					when "1000000001"=>	datao_i <= "0000" & reg_rominitend & reg_sdrddone & reg_sderr & reg_inuse;
					when "1000000010"=> datao_i <= "00" & reg_cmdno;
					when "1000000011"=> datao_i <= reg_sdadd( 7 downto  0);
					when "1000000100"=> datao_i <= reg_sdadd(15 downto  8);
					when "1000000101"=> datao_i <= reg_sdadd(23 downto 16);
					when "1000000110"=> datao_i <= reg_sdadd(31 downto 24);
					when "1000001100"=> datao_i <= FPGAVER( 7 downto  0);
					when "1000001101"=> datao_i <= FPGAVER(15 downto  8);
					when "1000001110"=> datao_i <= reg_firmver( 7 downto  0);
					when "1000001111"=> datao_i <= reg_firmver(15 downto  8);
					when "1000010000"=> datao_i <= "0000000" & reg_bufs;
					when "1000010001"=> datao_i <= "000000" & reg_bufinit & reg_bufrd;
					when "1000010010"=> datao_i <= "0000000" & reg_bufrdreq;
					when "1000010011"=> datao_i <= "000000" & reg_fifoinit & reg_fifordind;
					when "1000010100"=> datao_i <= "0000" & reg_cmtwr & reg_cmtrd & reg_cgrom & reg_sdram;
					when "1000010101"=> datao_i <= "000000" & reg_fifordend & reg_fifowrreq;
					when "1000010110"=> datao_i <= reg_romstart( 7 downto  0);
					when "1000010111"=> datao_i <= reg_romstart(15 downto  8);
					when "1000011000"=> datao_i <= reg_romstart(23 downto 16);
					when "1000011001"=> datao_i <= reg_romstart(31 downto 24);
					when "1000011010"=> datao_i <= reg_cmtlen( 7 downto  0);
					when "1000011011"=> datao_i <= reg_cmtlen(15 downto  8);
					when "1000011100"=> datao_i <= reg_cmtlen(23 downto 16);
					when "1000011101"=> datao_i <= reg_cmtlen(31 downto 24);
					when "1000011110"=> datao_i <= "000000" & cmtwropen_f2 & cmtrdopen_f2;
					when "1000100000"=> datao_i <= reg_cmtwnom( 7 downto  0);
					when "1000100001"=> datao_i <= reg_cmtwnom(15 downto  8);
					when "1000100010"=> datao_i <= reg_cmtwnom(23 downto 16);
					when "1000100011"=> datao_i <= reg_cmtwnom(31 downto 24);
					when "1000100100"=> datao_i <= reg_cmtwaccl( 7 downto  0);
					when "1000100101"=> datao_i <= reg_cmtwaccl(15 downto  8);
					when "1000100110"=> datao_i <= reg_cmtwaccl(23 downto 16);
					when "1000100111"=> datao_i <= reg_cmtwaccl(31 downto 24);
					when "1000101000"=> datao_i <= reg_cmtwacch( 7 downto  0);
					when "1000101001"=> datao_i <= reg_cmtwacch(15 downto  8);
					when "1000101010"=> datao_i <= reg_cmtwacch(23 downto 16);
					when "1000101011"=> datao_i <= reg_cmtwacch(31 downto 24);
					when "1000101100"=> datao_i <= reg_cmtwst( 7 downto  0);
					when "1000101101"=> datao_i <= reg_cmtwst(15 downto  8);
					when "1000101110"=> datao_i <= reg_cmtwst(23 downto 16);
					when "1000101111"=> datao_i <= reg_cmtwst(31 downto 24);
					when others      =>
						if (add_f2(9) = '0') then
							datao_i <= buf_q;
						else
							datao_i <= X"FF";
						end if;
				end case;
			end if;
		end if;
	end process;


-- SD read data counter
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			sd_cnt    <= (others => '1');
			reg_bufrdreq <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_bufinit = '1') then
				sd_cnt    <= (others => '0');
				reg_bufrdreq <= '0';
			elsif (indtenb_i = '1' and indtlt2_i = '1') then
				if (sd_cnt = "111111111") then
					reg_bufrdreq <= '1';
				else
					sd_cnt    <= sd_cnt + 1;
				end if;
			end if;
		end if;
	end process;


-- CMT read request latch
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			cmtrdrq_f1 <= '0';
			cmtrdrq_f2 <= '0';
			cmtrdrq_f3 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			cmtrdrq_f1 <= CMTRDREQ;
			cmtrdrq_f2 <= cmtrdrq_f1;
			cmtrdrq_f3 <= cmtrdrq_f2;
		end if;
	end process;

-- fifo address
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			fifo_cnt      <= (others => '0');
			fifo_add      <= (others => '0');
			reg_fifordend <= '0';
			reg_fifowrreq <= '0';
			cmtrdycnt     <= (others => '0');
			cmtrdy_i      <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_fifoinit = '1') then
				fifo_cnt      <= (others => '0');
				fifo_add      <= (others => '0');
				reg_fifordend <= '0';
				reg_fifowrreq <= '0';
				cmtrdycnt     <= (others => '0');
				cmtrdy_i      <= '0';
			elsif (reg_fifordind = '1') then
				if (fifo_add = reg_cmtlen) then
					reg_fifordend <= '1';
					reg_fifowrreq <= '0';
				elsif (reg_fifordend = '0' and reg_fifowrreq = '0') then

					if (reg_sdram = '1' or reg_cgrom = '1') then

						if (fifo_cnt = "11111") then
							fifo_cnt <= (others => '0');
							fifo_add <= fifo_add + 1;
							if (fifo_add(8 downto 0) = "111111111") then
								reg_fifowrreq <= '1';
							end if;
						else
							fifo_cnt <= fifo_cnt + 1;
						end if;

					elsif (reg_cmtrd = '1') then


						if (cmtrdrq_f2 = '1' and cmtrdrq_f3 = '0') then
							cmtrdycnt     <= (others => '0');
							cmtrdy_i      <= '0';
							fifo_add <= fifo_add + 1;
							if (fifo_add(8 downto 0) = "111111111") then
								reg_fifowrreq <= '1';
							end if;
						elsif (CMTRDACC = '0' and cmtrdycnt >= reg_cmtwst(20 downto 0) ) then
							cmtrdy_i      <= '1';
						elsif (CMTRDACC = '1' and CMTRDACCMD = '0' and cmtrdycnt >= reg_cmtwnom(20 downto 0) ) then
							cmtrdy_i      <= '1';
						elsif (CMTRDACC = '1' and CMTRDACCMD = '1' and cmtrdycnt >= reg_cmtwacch(20 downto 0) ) then
							cmtrdy_i      <= '1';
						else
							if (cmtrdopen_f2 = '0') then
								cmtrdycnt     <= (others => '0');
							else
								cmtrdycnt     <= cmtrdycnt + 1;
							end if;
							cmtrdy_i      <= '0';
						end if;

					end if;

				end if;
			else
				reg_fifowrreq <= '0';
			end if;
		end if;
	end process;


-- SDRAM/CGROM write data generate
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			reg_rominit_f1 <= '0';
			reg_rominit_f2 <= '0';
			reg_rominit_f3 <= '0';
		elsif (CLK50M'event and CLK50M = '1') then
			reg_rominit_f1 <= reg_rominit;
			reg_rominit_f2 <= reg_rominit_f1;
			reg_rominit_f3 <= reg_rominit_f2;
		end if;
	end process;

	reg_rominit_r <= '1' when (reg_rominit_f2 = '1' and reg_rominit_f3 = '0') else '0';


	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			sd_rdad_i  <= (others => '0');
			sd_rddt_i  <= (others => '0');
			sd_sdenb_i <= '0';
			sd_sdwrn_i <= '1';
			sd_cgenb_i <= '0';
			sd_cgwrn_i <= '1';
			rominitcnt <= (others => '0');
			reg_rominitend <= '0';
		elsif (CLK50M'event and CLK50M = '1') then

			if (reg_rominit_r = '1') then
				reg_rominitend <= '1';
			elsif (rominitcnt = "11" & X"FFFF") then
				reg_rominitend <= '0';
			end if;

			if (reg_rominitend = '1') then
				rominitcnt <= rominitcnt + 1;
			else
				rominitcnt <= (others => '0');
			end if;

			if (reg_rominitend = '1') then

				sd_rdad_i  <= "010" & "11" & rominitcnt(17 downto 4);
				sd_rddt_i  <= X"FF";
				sd_sdenb_i <= '1';

				if (rominitcnt(3 downto 2) = "10") then
					sd_sdwrn_i <= '0';
				else
					sd_sdwrn_i <= '1';
				end if;

			else

				sd_rdad_i <= ("000" & fifo_add(15 downto 0)) + reg_romstart(18 downto 0);

				if (reg_bufs = '1') then
					sd_rddt_i <= buf1_q;
				else
					sd_rddt_i <= buf2_q;
				end if;

				if (reg_fifordind = '1' and reg_sdram = '1') then
					sd_sdenb_i <= '1';
					if (fifo_cnt(4 downto 3) = "10") then
						sd_sdwrn_i <= '0';
					else
						sd_sdwrn_i <= '1';
					end if;
				else
					sd_sdenb_i <= '0';
					sd_sdwrn_i <= '1';
				end if;

			end if;

			if (reg_fifordind = '1' and reg_cgrom = '1') then
				sd_cgenb_i <= '1';
				if (fifo_cnt(4 downto 3) = "10") then
					sd_cgwrn_i <= '0';
				else
					sd_cgwrn_i <= '1';
				end if;
			else
				sd_cgenb_i <= '0';
				sd_cgwrn_i <= '1';
			end if;

		end if;
	end process;

-- CMT read data
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			cmtrddt_i <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			if (reg_bufs = '1') then
				cmtrddt_i <= buf1_q;
			else
				cmtrddt_i <= buf2_q;
			end if;
		end if;
	end process;


-- from 8049 data latch
	process (CLK50M,RSTN)
	begin
		if (RSTN = '0') then
			cmtrdopen_f1 <= '0';
			cmtrdopen_f2 <= '0';
			cmtrdreq_f1  <= '0';
			cmtrdreq_f2  <= '0';
			cmtwropen_f1 <= '0';
			cmtwropen_f2 <= '0';
			cmtwrreq_f1  <= '0';
			cmtwrreq_f2  <= '0';
			cmtwrdt_f1   <= (others => '0');
			cmtwrdt_f2   <= (others => '0');
		elsif (CLK50M'event and CLK50M = '1') then
			cmtrdopen_f1 <= CMTRDOPEN;
			cmtrdopen_f2 <= cmtrdopen_f1;
			cmtrdreq_f1  <= CMTRDREQ;
			cmtrdreq_f2  <= cmtrdreq_f1;
			cmtwropen_f1 <= CMTWROPEN;
			cmtwropen_f2 <= cmtwropen_f1;
			cmtwrreq_f1  <= CMTWRREQ;
			cmtwrreq_f2  <= cmtwrreq_f1;
			cmtwrdt_f1   <= CMTWRDT;
			cmtwrdt_f2   <= cmtwrdt_f1;
		end if;
	end process;


	DATAO     <= datao_i;

	SD_RDAD   <= sd_rdad_i;
	SD_RDDT   <= sd_rddt_i;
	SD_SDENB  <= sd_sdenb_i;
	SD_SDWRN  <= sd_sdwrn_i;
	SD_CGENB  <= sd_cgenb_i;
	SD_CGWRN  <= sd_cgwrn_i;

	SD_RDDONE <= reg_sdrddone;
	CMTRDY    <= cmtrdy_i;
	CMTRDDT   <= cmtrddt_i;

	SD_ERR    <= reg_sderr;
	SD_OUTENB <= reg_inuse;

	FIRMVER   <= reg_firmver;
	CMTCNT    <= fifo_add(15 downto 0);

end RTL;
